import numpy as np
import pickle


class History:

    def __init__(self, csvpath=None, warmup=0):
        '''
        warmup: `int`
            Write only if `self.count > warmup`.
            Useful if unknown columns appear after the first row.
        '''
        self.data = dict()  # Data columns.
        self.count = 0  # Current number of entries.
        self.warmup = warmup
        self.csvcount = 0  # Current number of entries written to CSV.
        self.csvpath = csvpath
        self.csvkeys = None
        self.csvfile = open(csvpath, 'w') if csvpath is not None else None

    @staticmethod
    def _none_like(value):
        if value is None:
            return None
        elif type(value) in [float, np.float32, np.float64]:
            return 0.
        elif type(value) == int:
            return 0
        else:
            assert "Unknown type: " + str(type(value))

    def append(self, key, value=None):
        assert value is None or isinstance(value, (
            int,
            float,
            str,
            np.float32,
            np.float64,
            np.ndarray,
        )), ("Unexpected type: " + str(type(value)))
        if isinstance(value, np.ndarray):
            assert value.shape == (1, ) or len(value.shape) == 0
            value = value.item()
        if key not in self.data:
            assert value is not None
            self.data[key] = [self._none_like(value)] * self.count
        if value is None:
            assert len(self.data[key]) > 0, 'Expected non-empty column ' + key
            value = self._none_like(self.data[key][-1])
        self.data[key].append(value)

    def commit(self):  # Finish the current entry.
        maxlen = max(len(self.data[k]) for k in self.data)
        fail = ''
        for k, v in self.data.items():
            if len(v) < maxlen:
                fail += k + ','
        if fail:
            raise RuntimeError('Missing values for columns: ' + fail)
        self.count += 1

    def get(self, key, default=None):
        return self.data.get(key, default)

    def append_dict(self, newdict):
        for k, v in newdict.items():
            self.append(k, v)

    def write(self, nocommit=False):
        '''
        Writes pending rows to current file.
        '''
        if not nocommit:
            self.commit()
        if self.count <= self.warmup:
            return
        if self.csvfile is None:
            return
        if self.csvkeys is not None:
            if len(self.data.keys()) != len(self.csvkeys):
                newkeys = list(set(self.data.keys()) - set(self.csvkeys))
                raise RuntimeError(
                    "Unexpected keys in history: {:}".format(newkeys))
        if self.csvcount == 0:
            # Write header.
            self.csvkeys = list(self.data.keys())
            self.csvfile.write(','.join(self.csvkeys) + '\n')
        while self.csvcount < self.count:
            row = [self.data[key][self.csvcount] for key in self.data]
            line = ','.join(map(str, row))
            self.csvfile.write(line + '\n')
            self.csvcount += 1
        self.csvfile.flush()

    def save(self, path):
        '''
        Saves data to pickle.
        '''
        with open(path, 'wb') as f:
            pickle.dump(self.data, f)

    def load(self, path):
        '''
        Overwrites data from pickle.
        '''
        with open(path, 'rb') as f:
            self.data = pickle.load(f)
            self.csvkeys = list(self.data.keys())
            self.count = len(next(iter(self.data.values())))
            self.write(nocommit=True)

    def close(self):
        self.csvfile.close()



import numpy as np


def solve(matr, rhs, args, status=None, linsolver="direct"):
    import scipy.sparse
    import scipy.sparse.linalg

    if args.linsolver_maxiter is None:
        if args.linsolver == 'lsqr':
            args.linsolver_maxiter = 1000
        else:
            args.linsolver_maxiter = 50

    def get_sparse_eye(size):
        return scipy.sparse.diags(np.ones(size), format='csr')

    eye = get_sparse_eye(matr.shape[1])
    matr_reg = matr.T.dot(matr).tocsr()
    if args.linsolver_damp:
        matr_reg += args.linsolver_damp**2 * eye
    if args.linsolver_dampdiag:
        matr_reg += args.linsolver_dampdiag**2 * scipy.sparse.diags(
            matr_reg.diagonal())
    rhs_reg = matr.T.dot(rhs)
    if linsolver == "direct":
        # Normal equations.
        sol = scipy.sparse.linalg.spsolve(matr_reg,
                                          rhs_reg,
                                          permc_spec='MMD_ATA')
    elif linsolver == "directsq":
        # Original system, assuming a square matrix.
        sol = scipy.sparse.linalg.spsolve(matr, rhs, permc_spec='MMD_ATA')
    elif linsolver == "direct_cu":
        import cupy
        import cupyx.scipy.sparse
        import cupyx.scipy.sparse.linalg
        matr_reg = cupyx.scipy.sparse.csr_matrix(matr_reg)
        rhs_reg = cupy.array(rhs_reg)
        sol = cupyx.scipy.sparse.linalg.spsolve(matr_reg, rhs_reg)
        sol = sol.get()
    elif linsolver == "sparseqr":
        import sparseqr
        sol = sparseqr.solve(matr, rhs, tolerance=args.linsolver_tol)
    elif linsolver == "lsqr":
        sol, _, itn, _, _, anorm, acond, arnorm = \
                scipy.sparse.linalg.lsqr(
            matr,
            rhs,
            damp=args.linsolver_damp,
            atol=args.linsolver_tol,
            btol=args.linsolver_tol,
            iter_lim=args.linsolver_maxiter)[:8]
        status['residual'] = arnorm
        status['anorm'] = anorm
        status['acond'] = acond
        status['niter'] = itn
    elif linsolver == "lsqr_cu":
        if cupy is None:
            raise ModuleNotFoundError(
                "Module CuPy not found. Install with 'pip install cupy-cuda110'"
            )
        # XXX cupy does not support non-square matrices
        sol = cupyx.scipy.sparse.linalg.lsqr(matr, rhs)[0]
    elif linsolver == "multigrid":
        import pyamg
        if pyamg is None:
            raise ModuleNotFoundError(
                "Module PyAMG not found. Install with 'pip install pyamg'")
        ml = pyamg.smoothed_aggregation_solver(matr_reg)
        residuals = []
        sol = ml.solve(b=rhs_reg,
                       tol=args.linsolver_tol,
                       residuals=residuals,
                       accel='cg',
                       maxiter=args.linsolver_maxiter)
        status['residual'] = residuals[-1]
        status['niter'] = len(residuals)
    elif linsolver == "bicgstab":
        residuals = []

        def callback(x):
            residuals.append(np.mean((matr_reg.dot(x) - rhs_reg)**2)**0.5)

        sol, _ = scipy.sparse.linalg.bicgstab(matr_reg,
                                              rhs_reg,
                                              tol=0,
                                              atol=args.linsolver_tol,
                                              callback=callback,
                                              maxiter=args.linsolver_maxiter)
        status['residual'] = residuals[-1]
        status['niter'] = len(residuals)
    else:
        raise ValueError("Unknown linsolver=" + linsolver)

    return sol


def add_arguments(parser):
    parser.add_argument('--linsolver',
                        type=str,
                        choices=[
                            "multigrid",
                            "direct",
                            "directsq",
                            "direct_cu",
                            "sparseqr",
                            "lsqr",
                            "lsqr_cu",
                            "bicgstab",
                        ],
                        default="direct",
                        help="Linear solver to use")
    parser.add_argument('--linsolver_maxiter',
                        type=int,
                        default=None,
                        help="Maximum number of iterations of linear solver")
    parser.add_argument('--linsolver_tol',
                        type=float,
                        default=1e-6,
                        help="Tolerance for linear solver")
    parser.add_argument('--linsolver_damp',
                        type=float,
                        default=0,
                        help="Relaxation factor (0: no relaxation)")
    parser.add_argument('--linsolver_dampdiag',
                        type=float,
                        default=0,
                        help="Multiplier for diagonal (0: no relaxation)")
    parser.add_argument('--linsolver_verbose',
                        type=int,
                        default=0,
                        help="Verbosity level for linsolver messages")
    parser.add_argument('--linsolver_history',
                        type=int,
                        default=0,
                        help="Dump history from linsolver status")
    parser.add_argument('--lr', type=float, default=1e-3, help="Learning rate")
    parser.add_argument('--nlvl',
                        type=int,
                        default=100,
                        help="Multigrid levels")
    parser.add_argument('--smooth_pre',
                        type=int,
                        default=2,
                        help="Pre-smoothing steps")
    parser.add_argument('--smooth_post',
                        type=int,
                        default=2,
                        help="Post-smoothing steps")
    parser.add_argument('--omega',
                        type=float,
                        default=0.6,
                        help="Jacobi smoother relaxation factor")
    parser.add_argument('--ndirect',
                        type=int,
                        default=3,
                        help="Systems on smaller grids "
                        "are solved with direct solver")
    parser.add_argument('--restriction',
                        type=str,
                        choices=('full', 'half', 'injection'),
                        default='full',
                        help="Multigrid restriction type")
